"""Models for OxaPay API."""

from .balance import AccountBalanceRequest, AccountBalanceResponse
from .base import OrderDirection, PaginatedResponse, ResultResponse, SortField
from .exchange import (
    ExchangeCalculateRequest,
    ExchangeCalculateResponse,
    ExchangeHistoryRequest,
    ExchangeHistoryResponse,
    ExchangePairsResponse,
    ExchangeRateRequest,
    ExchangeRateResponse,
    ExchangeRequestRequest,
    ExchangeRequestResponse,
    ExchangeType,
)
from .merchant import (
    AcceptedCoinsRequest,
    AcceptedCoinsResponse,
    CreateInvoiceRequest,
    CreateInvoiceResponse,
    CreateStaticWalletRequest,
    CreateStaticWalletResponse,
    CreateWhiteLabelRequest,
    CreateWhiteLabelResponse,
    PaymentHistoryRequest,
    PaymentHistoryResponse,
    PaymentInfoRequest,
    PaymentInfoResponse,
    PaymentStatus,
    PaymentType,
    RevokeStaticWalletRequest,
    RevokeStaticWalletResponse,
    StaticWalletListRequest,
    StaticWalletListResponse,
)
from .payout import (
    CreatePayoutRequest,
    CreatePayoutResponse,
    PayoutHistoryRequest,
    PayoutHistoryResponse,
    PayoutInfoRequest,
    PayoutInfoResponse,
    PayoutStatus,
    PayoutType,
)
from .system import (
    PriceResponse,
    SupportedCurrenciesResponse,
    SupportedFiatCurrenciesResponse,
    SupportedNetworksResponse,
)

__all__ = [  # noqa: RUF022
    # Base models
    "ResultResponse",
    "PaginatedResponse",
    "OrderDirection",
    "SortField",
    # Merchant models
    "CreateInvoiceRequest",
    "CreateInvoiceResponse",
    "CreateWhiteLabelRequest",
    "CreateWhiteLabelResponse",
    "CreateStaticWalletRequest",
    "CreateStaticWalletResponse",
    "RevokeStaticWalletRequest",
    "RevokeStaticWalletResponse",
    "StaticWalletListRequest",
    "StaticWalletListResponse",
    "PaymentInfoRequest",
    "PaymentInfoResponse",
    "PaymentHistoryRequest",
    "PaymentHistoryResponse",
    "AcceptedCoinsRequest",
    "AcceptedCoinsResponse",
    "PaymentStatus",
    "PaymentType",
    # Payout models
    "CreatePayoutRequest",
    "CreatePayoutResponse",
    "PayoutInfoRequest",
    "PayoutInfoResponse",
    "PayoutHistoryRequest",
    "PayoutHistoryResponse",
    "PayoutStatus",
    "PayoutType",
    # Balance models
    "AccountBalanceRequest",
    "AccountBalanceResponse",
    # Exchange models
    "ExchangeRateRequest",
    "ExchangeRateResponse",
    "ExchangeCalculateRequest",
    "ExchangeCalculateResponse",
    "ExchangePairsResponse",
    "ExchangeRequestRequest",
    "ExchangeRequestResponse",
    "ExchangeHistoryRequest",
    "ExchangeHistoryResponse",
    "ExchangeType",
    # System models
    "SupportedCurrenciesResponse",
    "SupportedFiatCurrenciesResponse",
    "SupportedNetworksResponse",
    "PriceResponse",
]



"""FastAPI integration for OxaPay webhook handling."""

import hashlib
import hmac
import json
import logging
from collections.abc import Callable
from collections.abc import Callable as CallableType
from datetime import datetime
from typing import Annotated, Any, TypeVar

from fastapi import FastAPI, HTTPException, Request
from pydantic import BaseModel, Field

T = TypeVar("T", bound=BaseModel)

logger = logging.getLogger(__name__)


class WebhookEvent(BaseModel):
    """Base model for webhook events."""

    status: str
    track_id: Annotated[str, Field(validation_alias="trackId")]
    type: str


class PaymentWebhookEvent(WebhookEvent):
    """Model for payment webhook events."""

    amount: float
    currency: str
    fee_paid_by_payer: Annotated[bool, Field(validation_alias="feePaidByPayer")]
    under_paid_cover: Annotated[bool, Field(validation_alias="underPaidCover")]
    email: Annotated[str | None, Field(None, validation_alias="email")]
    order_id: Annotated[str | None, Field(None, validation_alias="orderId")]
    description: Annotated[str | None, Field(None)]
    date: datetime
    pay_date: Annotated[datetime | None, Field(None, validation_alias="payDate")]
    address: Annotated[str | None, Field(None)]
    sender_address: Annotated[str | None, Field(None, validation_alias="senderAddress")]
    tx_id: Annotated[str | None, Field(None, validation_alias="txID")]
    price: Annotated[str | None, Field(None)]
    pay_amount: Annotated[str | None, Field(None, validation_alias="payAmount")]
    pay_currency: Annotated[str | None, Field(None, validation_alias="payCurrency")]
    network: Annotated[str | None, Field(None)]

    class Config:  # noqa: D106
        json_encoders = {datetime: lambda dt: dt.isoformat()}  # noqa: RUF012


class PayoutWebhookEvent(WebhookEvent):
    """Model for payout webhook events."""

    tx_id: Annotated[str | None, Field(None, validation_alias="txID")]
    address: str
    amount: float
    currency: str
    price: Annotated[float | None, Field(None)]
    network: str
    date: datetime
    description: Annotated[str | None, Field(None)]

    class Config:  # noqa: D106
        json_encoders = {datetime: lambda dt: dt.isoformat()}  # noqa: RUF012


WebhookEventType = PaymentWebhookEvent | PayoutWebhookEvent


class WebhookRouter:
    """A class for handling OxaPay webhooks with custom callbacks.

    This class allows setting up callbacks for different webhook events.
    """

    def __init__(
        self,
        app: FastAPI,
        merchant_api_key: str | None = None,
        payout_api_key: str | None = None,
        path: str = "/oxapay/webhook",
    ) -> None:
        """Initialize the webhook router.

        Args:
            app: The FastAPI application
            merchant_api_key: API key for merchant operations (payment webhooks)
            payout_api_key: API key for payout operations (payout webhooks)
            path: The path where the webhook will be set up

        """
        self.app = app
        self.merchant_api_key = merchant_api_key
        self.payout_api_key = payout_api_key
        self.path = path

        # Callbacks for different webhook events
        self.payment_waiting_callback = None
        self.payment_confirming_callback = None
        self.payment_paid_callback = None
        self.payment_expired_callback = None
        self.payment_failed_callback = None

        self.payout_confirming_callback = None
        self.payout_complete_callback = None
        self.payout_rejected_callback = None

        # Set up the webhook handler
        self._setup_webhook_handler()

    def on_payment_waiting(
        self,
        callback: Callable[[PaymentWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payment waiting events.

        Args:
            callback: Function to call when a payment waiting event is received

        Returns:
            The callback function

        """
        self.payment_waiting_callback = callback
        return callback

    def on_payment_confirming(
        self,
        callback: Callable[[PaymentWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payment confirming events.

        Args:
            callback: Function to call when a payment confirming event is received

        Returns:
            The callback function

        """
        self.payment_confirming_callback = callback
        return callback

    def on_payment_paid(
        self,
        callback: Callable[[PaymentWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payment paid events.

        Args:
            callback: Function to call when a payment paid event is received

        Returns:
            The callback function

        """
        self.payment_paid_callback = callback
        return callback

    def on_payment_expired(
        self,
        callback: Callable[[PaymentWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payment expired events.

        Args:
            callback: Function to call when a payment expired event is received

        Returns:
            The callback function

        """
        self.payment_expired_callback = callback
        return callback

    def on_payment_failed(
        self,
        callback: Callable[[PaymentWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payment failed events.

        Args:
            callback: Function to call when a payment failed event is received

        Returns:
            The callback function

        """
        self.payment_failed_callback = callback
        return callback

    def on_payout_confirming(
        self,
        callback: Callable[[PayoutWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payout confirming events.

        Args:
            callback: Function to call when a payout confirming event is received

        Returns:
            The callback function

        """
        self.payout_confirming_callback = callback
        return callback

    def on_payout_complete(
        self,
        callback: Callable[[PayoutWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payout complete events.

        Args:
            callback: Function to call when a payout complete event is received

        Returns:
            The callback function

        """
        self.payout_complete_callback = callback
        return callback

    def on_payout_rejected(
        self,
        callback: Callable[[PayoutWebhookEvent], Any],
    ) -> Callable:
        """Register a callback for payout rejected events.

        Args:
            callback: Function to call when a payout rejected event is received

        Returns:
            The callback function

        """
        self.payout_rejected_callback = callback
        return callback

    def _setup_webhook_handler(self) -> None:
        @self.app.post(self.path)
        async def webhook_handler(request: Request) -> dict[str, str]:
            return await self._process_webhook_request(request)

    def _verify_signature(self, api_key: str, request: Request) -> bool:
        if not api_key:
            return False

        hmac_header = request.headers.get("HMAC")
        if not hmac_header:
            return False

        payload = request.scope["body"].decode("utf-8")
        calculated_hmac = hmac.new(
            api_key.encode(),
            payload.encode(),
            hashlib.sha512,
        ).hexdigest()

        return calculated_hmac == hmac_header

    def _raise_api_key_error(self, event_type: str) -> None:
        detail = (
            "Merchant API key not configured"
            if event_type == "payment"
            else "Payout API key not configured"
        )
        raise HTTPException(status_code=500, detail=detail)

    def _raise_unknown_event_type(self, event_type: str) -> None:
        raise HTTPException(
            status_code=400,
            detail=f"Unknown event type: {event_type}",
        )

    def _raise_invalid_signature(self) -> None:
        raise HTTPException(status_code=401, detail="Invalid signature")

    def _parse_webhook_data(
        self,
        data: dict[str, Any],
        event_model: type[WebhookEventType],
    ) -> WebhookEventType:
        try:
            return event_model.parse_obj(data)
        except Exception as e:
            raise HTTPException(
                status_code=400,
                detail=f"Invalid webhook payload: {e!s}",
            ) from e

    def _handle_payment_event(self, event: PaymentWebhookEvent) -> None:
        handlers = {
            "Waiting": self.payment_waiting_callback,
            "Confirming": self.payment_confirming_callback,
            "Paid": self.payment_paid_callback,
            "Expired": self.payment_expired_callback,
            "Failed": self.payment_failed_callback,
        }

        if callback := handlers.get(event.status):
            self.call_callback_wrapper(callback, event)

    def _handle_payout_event(self, event: PayoutWebhookEvent) -> None:
        handlers = {
            "Confirming": self.payout_confirming_callback,
            "Complete": self.payout_complete_callback,
            "Rejected": self.payout_rejected_callback,
        }

        if callback := handlers.get(event.status):
            self.call_callback_wrapper(callback, event)

    async def _process_webhook_request(
        self,
        request: Request,
    ) -> dict[str, str]:
        # Read and parse the raw request body
        body = await request.body()
        request.scope["body"] = body  # Store the raw body for signature verification

        try:
            data = json.loads(body)
            event_type = data.get("type")

            # Determine API key and event model based on event type
            if event_type == "payment":
                if not self.merchant_api_key:
                    self._raise_api_key_error(event_type)
                api_key = self.merchant_api_key
                event_model = PaymentWebhookEvent
                handler = self._handle_payment_event
            elif event_type == "payout":
                if not self.payout_api_key:
                    self._raise_api_key_error(event_type)
                api_key = self.payout_api_key
                event_model = PayoutWebhookEvent
                handler = self._handle_payout_event
            else:
                self._raise_unknown_event_type(event_type)

            # Verify the webhook signature
            if not self._verify_signature(api_key, request):
                self._raise_invalid_signature()

            # Parse and validate the webhook data
            event = self._parse_webhook_data(data, event_model)

            # Handle the event
            handler(event)

            # Return a success response
            return {"status": "ok"}  # noqa: TRY300
        except json.JSONDecodeError as err:
            # Print error
            raise HTTPException(
                status_code=400,
                detail="Invalid JSON payload",
            ) from err
        except HTTPException:
            # Re-raise HTTPExceptions without modification
            raise
        except Exception as e:
            logger.exception("Error processing webhook")
            raise HTTPException(
                status_code=500,
                detail=f"Error processing webhook: {e!s}",
            ) from e

    async def _call_callback(
        self,
        callback: CallableType,
        event: WebhookEventType,
    ) -> None:
        try:
            import inspect

            if inspect.iscoroutinefunction(callback):
                await callback(event)
            else:
                callback(event)
        except Exception:
            # Log the error but don't fail the webhook response
            logger.exception("Error in webhook callback")

    def call_callback_wrapper(
        self,
        callback: CallableType,
        event: WebhookEventType,
    ) -> None:
        """Handle calling callbacks in both sync and async contexts.

        Args:
            callback: The callback function to call
            event: The webhook event to pass to the callback

        """
        import asyncio

        try:
            asyncio.create_task(self._call_callback(callback, event))  # noqa: RUF006
        except RuntimeError:
            # We're not in an async context, create a new event loop
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self._call_callback(callback, event))
            loop.close()



"""Main client implementation for OxaPay API."""

import hashlib
import hmac
import types
from typing import Any, TypeVar

import aiohttp
from pydantic import BaseModel

from .exceptions import OxaPayError, OxaPayValidationError
from .models.balance import AccountBalanceRequest, AccountBalanceResponse
from .models.exchange import (
    ExchangeCalculateRequest,
    ExchangeCalculateResponse,
    ExchangeHistoryRequest,
    ExchangeHistoryResponse,
    ExchangePairsResponse,
    ExchangeRateRequest,
    ExchangeRateResponse,
    ExchangeRequestRequest,
    ExchangeRequestResponse,
)
from .models.merchant import (
    AcceptedCoinsResponse,
    CreateInvoiceRequest,
    CreateInvoiceResponse,
    CreateStaticWalletRequest,
    CreateStaticWalletResponse,
    CreateWhiteLabelRequest,
    CreateWhiteLabelResponse,
    PaymentHistoryRequest,
    PaymentHistoryResponse,
    PaymentInfoRequest,
    PaymentInfoResponse,
    RevokeStaticWalletRequest,
    RevokeStaticWalletResponse,
    StaticWalletListRequest,
    StaticWalletListResponse,
)
from .models.payout import (
    CreatePayoutRequest,
    CreatePayoutResponse,
    PayoutHistoryRequest,
    PayoutHistoryResponse,
    PayoutInfoRequest,
    PayoutInfoResponse,
)
from .models.system import (
    PriceResponse,
    SupportedCurrenciesResponse,
    SupportedFiatCurrenciesResponse,
    SupportedNetworksResponse,
)

T = TypeVar("T", bound=BaseModel)


class OxaPayClient:
    """Client for interacting with the OxaPay API.

    Provides methods for all OxaPay API endpoints with proper typing and error handling.
    """

    BASE_URL = "https://api.oxapay.com"
    SUCCESS_RESULT_CODE = 100  # Success result code from API

    def __init__(
        self,
        merchant_api_key: str | None = None,
        payout_api_key: str | None = None,
        general_api_key: str | None = None,
    ) -> None:
        """Initialize the OxaPay client.

        Args:
            merchant_api_key: API key for merchant operations
            payout_api_key: API key for payout operations
            general_api_key: API key for general operations like exchange

        """
        self.merchant_api_key = merchant_api_key
        self.payout_api_key = payout_api_key
        self.general_api_key = general_api_key
        self.session = None

    async def __aenter__(self) -> "OxaPayClient":
        """Enter async context manager.

        Returns:
            The OxaPayClient instance.

        """
        self.session = aiohttp.ClientSession()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Exit async context manager.

        Args:
            exc_type: Exception type if an exception was raised
            exc_val: Exception value if an exception was raised
            exc_tb: Exception traceback if an exception was raised

        """
        if self.session:
            await self.session.close()
            self.session = None

    async def _request(
        self,
        method: str,
        endpoint: str,
        data: dict[str, Any] | None = None,
        response_model: type[T] | None = None,
    ) -> T | dict[str, Any]:
        """Make a request to the OxaPay API.

        Args:
            method: HTTP method (GET, POST)
            endpoint: API endpoint path
            data: Request data
            response_model: Pydantic model for response validation

        Returns:
            Validated response data or raw response

        Raises:
            OxaPayError: For API errors
            OxaPayValidationError: For validation errors

        """
        url = f"{self.BASE_URL}{endpoint}"

        if self.session is None:
            self.session = aiohttp.ClientSession()
            close_after = True
        else:
            close_after = False

        try:
            async with self.session.request(method, url, json=data) as response:
                response_data = await response.json()

                # Check for API errors
                if (
                    "result" in response_data
                    and response_data["result"] != self.SUCCESS_RESULT_CODE
                ):
                    raise OxaPayError(
                        code=response_data.get("result", 0),
                        message=response_data.get("message", "Unknown error"),
                    )

                # Validate and return the response
                if response_model:
                    try:
                        return response_model.parse_obj(response_data)
                    except Exception as e:
                        msg = f"Response validation error: {e!s}"
                        raise OxaPayValidationError(msg) from e
                return response_data
        finally:
            if close_after:
                await self.session.close()
                self.session = None

    # Merchant API Methods

    async def create_invoice(
        self,
        request: CreateInvoiceRequest,
    ) -> CreateInvoiceResponse:
        """Create a new payment link.

        Args:
            request: Invoice creation parameters

        Returns:
            Invoice creation response with payment link and tracking ID

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/request",
            data=request_data,
            response_model=CreateInvoiceResponse,
        )

    async def create_white_label_payment(
        self,
        request: CreateWhiteLabelRequest,
    ) -> CreateWhiteLabelResponse:
        """Create a white-labeled payment.

        Args:
            request: White-label payment creation parameters

        Returns:
            White-label payment response with payment details

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/request/whitelabel",
            data=request_data,
            response_model=CreateWhiteLabelResponse,
        )

    async def create_static_wallet(
        self,
        request: CreateStaticWalletRequest,
    ) -> CreateStaticWalletResponse:
        """Create a static wallet address.

        Args:
            request: Static wallet creation parameters

        Returns:
            Static wallet address creation response

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/request/staticaddress",
            data=request_data,
            response_model=CreateStaticWalletResponse,
        )

    async def revoke_static_wallet(
        self,
        request: RevokeStaticWalletRequest,
    ) -> RevokeStaticWalletResponse:
        """Revoke a static wallet address.

        Args:
            request: Static wallet revocation parameters

        Returns:
            Static wallet revocation response

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/revoke/staticaddress",
            data=request_data,
            response_model=RevokeStaticWalletResponse,
        )

    async def get_static_wallets(
        self,
        request: StaticWalletListRequest,
    ) -> StaticWalletListResponse:
        """Get a list of static wallets.

        Args:
            request: Static wallet list request parameters

        Returns:
            List of static wallets with pagination info

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/list/staticaddress",
            data=request_data,
            response_model=StaticWalletListResponse,
        )

    async def get_payment_info(self, track_id: str | int) -> PaymentInfoResponse:
        """Get information about a specific payment.

        Args:
            track_id: The payment tracking ID

        Returns:
            Payment information

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request = PaymentInfoRequest(trackId=str(track_id))
        request_data = request.model_dump(by_alias=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/inquiry",
            data=request_data,
            response_model=PaymentInfoResponse,
        )

    async def get_payment_history(
        self,
        request: PaymentHistoryRequest,
    ) -> PaymentHistoryResponse:
        """Get payment history.

        Args:
            request: Payment history request parameters

        Returns:
            Payment history with pagination info

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["merchant"] = self.merchant_api_key

        return await self._request(
            "POST",
            "/merchants/list",
            data=request_data,
            response_model=PaymentHistoryResponse,
        )

    async def get_accepted_coins(self) -> AcceptedCoinsResponse:
        """Get a list of accepted cryptocurrencies.

        Returns:
            List of accepted cryptocurrency symbols

        """
        if not self.merchant_api_key:
            msg = "Merchant API key is required"
            raise ValueError(msg)

        request_data = {"merchant": self.merchant_api_key}

        return await self._request(
            "POST",
            "/merchants/allowedCoins",
            data=request_data,
            response_model=AcceptedCoinsResponse,
        )

    # Payout API Methods

    async def create_payout(self, request: CreatePayoutRequest) -> CreatePayoutResponse:
        """Create a new payout.

        Args:
            request: Payout creation parameters

        Returns:
            Payout creation response with tracking ID

        """
        if not self.payout_api_key:
            msg = "Payout API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["key"] = self.payout_api_key

        return await self._request(
            "POST",
            "/api/send",
            data=request_data,
            response_model=CreatePayoutResponse,
        )

    async def get_payout_info(self, track_id: str | int) -> PayoutInfoResponse:
        """Get information about a specific payout.

        Args:
            track_id: The payout tracking ID

        Returns:
            Payout information

        """
        if not self.payout_api_key:
            msg = "Payout API key is required"
            raise ValueError(msg)

        request = PayoutInfoRequest(trackId=str(track_id))
        request_data = request.model_dump(by_alias=True)
        request_data["key"] = self.payout_api_key

        return await self._request(
            "POST",
            "/api/inquiry",
            data=request_data,
            response_model=PayoutInfoResponse,
        )

    async def get_payout_history(
        self,
        request: PayoutHistoryRequest,
    ) -> PayoutHistoryResponse:
        """Get payout history.

        Args:
            request: Payout history request parameters

        Returns:
            Payout history with pagination info

        """
        if not self.payout_api_key:
            msg = "Payout API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["key"] = self.payout_api_key

        return await self._request(
            "POST",
            "/api/list",
            data=request_data,
            response_model=PayoutHistoryResponse,
        )

    async def get_account_balance(
        self,
        currency: str | None = None,
    ) -> AccountBalanceResponse:
        """Get account balance for all or a specific currency.

        Args:
            currency: Specific currency to get balance for (optional)

        Returns:
            Account balance information

        """
        if not self.payout_api_key:
            msg = "Payout API key is required"
            raise ValueError(msg)

        request = AccountBalanceRequest(currency=currency)
        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["key"] = self.payout_api_key

        return await self._request(
            "POST",
            "/api/balance",
            data=request_data,
            response_model=AccountBalanceResponse,
        )

    # Exchange API Methods

    async def get_exchange_rate(
        self,
        request: ExchangeRateRequest,
    ) -> ExchangeRateResponse:
        """Get exchange rate between two currencies.

        Args:
            request: Exchange rate request parameters

        Returns:
            Exchange rate information

        """
        request_data = request.model_dump(by_alias=True)

        return await self._request(
            "POST",
            "/exchange/rate",
            data=request_data,
            response_model=ExchangeRateResponse,
        )

    async def calculate_exchange(
        self,
        request: ExchangeCalculateRequest,
    ) -> ExchangeCalculateResponse:
        """Calculate exchange amount.

        Args:
            request: Exchange calculation request parameters

        Returns:
            Exchange calculation result

        """
        request_data = request.model_dump(by_alias=True)

        return await self._request(
            "POST",
            "/exchange/calculate",
            data=request_data,
            response_model=ExchangeCalculateResponse,
        )

    async def get_exchange_pairs(self) -> ExchangePairsResponse:
        """Get available exchange pairs.

        Returns:
            List of available exchange pairs

        """
        return await self._request(
            "POST",
            "/exchange/pairs",
            response_model=ExchangePairsResponse,
        )

    async def create_exchange_request(
        self,
        request: ExchangeRequestRequest,
    ) -> ExchangeRequestResponse:
        """Create a new exchange request.

        Args:
            request: Exchange request parameters

        Returns:
            Exchange request creation response

        """
        if not self.general_api_key:
            msg = "General API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["key"] = self.general_api_key

        return await self._request(
            "POST",
            "/exchange/request",
            data=request_data,
            response_model=ExchangeRequestResponse,
        )

    async def get_exchange_history(
        self,
        request: ExchangeHistoryRequest,
    ) -> ExchangeHistoryResponse:
        """Get exchange history.

        Args:
            request: Exchange history request parameters

        Returns:
            Exchange history with pagination info

        """
        if not self.general_api_key:
            msg = "General API key is required"
            raise ValueError(msg)

        request_data = request.model_dump(by_alias=True, exclude_none=True)
        request_data["key"] = self.general_api_key

        return await self._request(
            "POST",
            "/exchange/list",
            data=request_data,
            response_model=ExchangeHistoryResponse,
        )

    # System API Methods

    async def get_prices(self) -> PriceResponse:
        """Get current prices of all supported cryptocurrencies.

        Returns:
            Current prices information

        """
        return await self._request("POST", "/api/prices", response_model=PriceResponse)

    async def get_supported_currencies(self) -> SupportedCurrenciesResponse:
        """Get a list of supported currencies.

        Returns:
            List of supported currencies with network details

        """
        return await self._request(
            "POST",
            "/api/currencies",
            response_model=SupportedCurrenciesResponse,
        )

    async def get_supported_fiat_currencies(self) -> SupportedFiatCurrenciesResponse:
        """Get a list of supported fiat currencies.

        Returns:
            List of supported fiat currencies

        """
        return await self._request(
            "POST",
            "/api/fiats",
            response_model=SupportedFiatCurrenciesResponse,
        )

    async def get_supported_networks(self) -> SupportedNetworksResponse:
        """Get a list of supported blockchain networks.

        Returns:
            List of supported networks

        """
        return await self._request(
            "POST",
            "/api/networks",
            response_model=SupportedNetworksResponse,
        )

    async def check_system_status(self) -> str:
        """Check if the OxaPay API is operational.

        Returns:
            Status message (usually "OK")

        """
        response = await self._request("POST", "/monitor")
        return (
            "OK"
            if isinstance(response, dict) and response.get("status") == "ok"
            else str(response)
        )

    def verify_webhook_signature(
        self,
        signature: str,
        payload: str,
        *,
        is_payment: bool = True,
    ) -> bool:
        """Verify the HMAC signature of a webhook payload.

        Args:
            signature: The HMAC signature from the webhook header
            payload: The raw webhook payload
            is_payment: Whether this is a payment webhook (True) or payout webhook (False)

        Returns:
            True if the signature is valid, False otherwise

        """  # noqa: E501
        api_key = self.merchant_api_key if is_payment else self.payout_api_key
        if not api_key:
            key_type = "Merchant" if is_payment else "Payout"
            msg = f"{key_type} API key is required"
            raise ValueError(msg)

        calculated_hmac = hmac.new(
            api_key.encode(),
            payload.encode(),
            hashlib.sha512,
        ).hexdigest()

        return calculated_hmac == signature



"""OxaPay Python Client.

A fully-typed client for interacting with the OxaPay API.
"""

from .client import OxaPayClient
from .exceptions import OxaPayError, OxaPayValidationError
from .models import *  # noqa: F403

__version__ = "0.2.1"
__all__ = [
    "OxaPayClient",
    "OxaPayError",
    "OxaPayValidationError",
]



"""Base models for OxaPay API."""

from datetime import datetime
from enum import Enum
from typing import Any, Generic, TypeVar

from pydantic import BaseModel, Field, model_validator

T = TypeVar("T")


def timestamp_to_datetime(value: int | str | None) -> datetime | None:
    """Convert a Unix timestamp to a datetime object."""
    if value is None:
        return None
    # Convert string to int if needed
    if isinstance(value, str) and value.isdigit():
        value = int(value)
    # Handle timestamp as int
    if isinstance(value, int | float):
        return datetime.fromtimestamp(value)  # noqa: DTZ006
    return value  # Return as is if already a datetime or incompatible format


class TimestampModelMixin:
    """Mixin for models with timestamp fields."""

    @model_validator(mode="before")
    @classmethod
    def convert_timestamps(cls, data: Any) -> Any:  # noqa: ANN401
        """Convert all timestamp fields to datetime objects."""
        if not isinstance(data, dict):
            return data

        # Standard timestamp field names to convert
        timestamp_fields = ["date", "payDate", "createdAt", "expiredAt"]

        # Process known timestamp fields
        for field in timestamp_fields:
            if field in data and data[field] is not None:
                data[field] = timestamp_to_datetime(data[field])

        # Process fields ending with 'Date' or 'At'
        for field in list(data.keys()):
            if (
                (field.endswith(("Date", "At")))
                and field not in timestamp_fields
                and data[field] is not None
            ):
                data[field] = timestamp_to_datetime(data[field])

        return data


class ResultResponse(BaseModel):
    """Base response with result code and message."""

    result: int = Field(
        ...,
        description="The result code indicates the success or failure of the request",
    )
    message: str = Field(
        ...,
        description="A message providing additional information about the result",
    )


class PaginatedResponse(ResultResponse, Generic[T]):
    """Base model for paginated responses."""

    data: list[T] = Field(..., description="The paginated data")
    meta: dict[str, int] = Field(
        ...,
        description="Pagination metadata (size, page, pages, total)",
    )


class OrderDirection(str, Enum):
    """Order direction for listings."""

    ASC = "asc"
    DESC = "desc"


class SortField(str, Enum):
    """Sort field options for listings."""

    CREATE_DATE = "create_date"
    PAY_DATE = "pay_date"
    AMOUNT = "amount"



import asyncio

from oxapay import OxaPayClient
from oxapay.models import CreateInvoiceRequest


async def create_invoice() -> None:
    """Example of creating a payment invoice."""
    # Replace with your actual API key
    merchant_api_key = "your_merchant_api_key"

    async with OxaPayClient(merchant_api_key=merchant_api_key) as client:
        # Create invoice request
        request = CreateInvoiceRequest(
            amount=100.50,
            currency="USD",  # Invoice in USD
            order_id="example-order-123",
            description="Purchase of premium widgets",
            email="customer@example.com",
            callback_url="https://your-site.com/oxapay-callback",
            return_url="https://your-site.com/thank-you",
            life_time=60,  # 60 minutes expiration
            fee_paid_by_payer=1,  # Customer pays the transaction fee
        )

        # Send the request
        try:
            response = await client.create_invoice(request)
            print(f"Payment Link: {response.pay_link}")
            print(f"Track ID: {response.track_id}")

            # You would typically store the track_id in your database
            # to associate it with the customer's order

        except Exception as e:
            print(f"Error creating invoice: {e}")


if __name__ == "__main__":
    asyncio.run(create_invoice())



"""Models for OxaPay Balance API."""

from pydantic import BaseModel, Field

from .base import ResultResponse


class AccountBalanceRequest(BaseModel):
    """Request model for account balance."""

    currency: str | None = Field(
        None,
        description="Specific currency to get balance for",
    )


class AccountBalanceResponse(ResultResponse):
    """Response model for account balance."""

    data: dict[str, str] = Field(..., description="Balances for each currency")



"""Exceptions for the OxaPay client."""


class OxaPayError(Exception):
    """Exception raised for OxaPay API errors.

    Attributes:
        code: The error code from the API
        message: The error message from the API

    """

    def __init__(self, code: int, message: str) -> None:
        """Initialize OxaPayError with error code and message.

        Args:
            code: The error code from the API
            message: The error message from the API

        """
        self.code = code
        self.message = message
        super().__init__(f"OxaPay API Error ({code}): {message}")


class OxaPayValidationError(Exception):
    """Exception raised for data validation errors."""



import asyncio
import sys

from oxapay import OxaPayClient, OxaPayError


async def check_payment_status(track_id: str) -> None:
    """Example of checking a payment's status."""
    # Replace with your actual API key
    merchant_api_key = "your_merchant_api_key"

    async with OxaPayClient(merchant_api_key=merchant_api_key) as client:
        try:
            # Get payment information
            payment = await client.get_payment_info(track_id)

            print(f"Payment {track_id} information:")
            print(f"Status: {payment.status}")
            print(f"Amount: {payment.amount} {payment.currency}")

            if payment.status == "Paid":
                print(f"Payment completed at: {payment.pay_date}")
                print(f"Transaction ID: {payment.tx_id}")
            elif payment.status == "Expired":
                print("Payment has expired")
            elif payment.status in ["New", "Waiting", "Confirming"]:
                print("Payment is still in progress")

        except OxaPayError as e:
            print(f"API Error {e.code}: {e.message}")
        except Exception as e:
            print(f"Error checking payment: {e}")


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Usage: python check_payment_status.py <track_id>")
        sys.exit(1)

    track_id = sys.argv[1]
    asyncio.run(check_payment_status(track_id))



import uvicorn
from fastapi import FastAPI

from oxapay.models import PaymentWebhookEvent
from oxapay.webhook import WebhookRouter

# Create FastAPI app
app = FastAPI(title="OxaPay Webhook Example")

# Initialize the webhook router
webhook = WebhookRouter(
    app=app,
    merchant_api_key="your_merchant_api_key",  # Replace with actual key
    payout_api_key="your_payout_api_key",  # Replace with actual key
    path="/oxapay/webhook",
)


# Register payment webhook handlers
@webhook.on_payment_waiting
async def handle_payment_waiting(event: PaymentWebhookEvent) -> None:
    """Handle payments that are waiting for confirmation."""
    print(f"Payment {event.track_id} is waiting for confirmation")
    print(f"Amount: {event.amount} {event.currency}")
    print(f"Customer email: {event.email}")
    # Update order status in your system


@webhook.on_payment_confirming
async def handle_payment_confirming(event: PaymentWebhookEvent) -> None:
    """Handle payments that are being confirmed on the blockchain."""
    print(f"Payment {event.track_id} is confirming on the blockchain")
    print(f"Transaction ID: {event.tx_id}")
    # Update order status in your system


@webhook.on_payment_paid
async def handle_payment_paid(event: PaymentWebhookEvent) -> None:
    """Handle completed payments."""
    print(f"Payment {event.track_id} has been completed!")
    print(f"Amount: {event.amount} {event.currency}")
    print(f"Paid with: {event.pay_amount} {event.pay_currency}")

    # Here you would typically:
    # 1. Verify the payment matches an order in your system
    # 2. Mark the order as paid
    # 3. Trigger fulfillment process

    # Example database update (pseudo-code):
    # db.update_order(order_id=event.order_id, status="paid", payment_id=event.track_id)

    print(f"Order {event.order_id} has been marked as paid and is being fulfilled")


@webhook.on_payment_expired
def handle_payment_expired(event: PaymentWebhookEvent) -> None:
    """Handle expired payments."""
    print(f"Payment {event.track_id} has expired")
    # Update order status in your system to reflect failed payment
    # Maybe send an email to the customer


@webhook.on_payment_failed
def handle_payment_failed(event: PaymentWebhookEvent) -> None:
    """Handle failed payments."""
    print(f"Payment {event.track_id} has failed")
    print("Reason: Payment processing failed")
    # Update order status and possibly notify customer


# Add a simple status endpoint
@app.get("/status")
def status() -> dict[str, str]:
    """Check if the server is running."""
    return {"status": "online"}


if __name__ == "__main__":
    # Run the webhook server
    uvicorn.run(app, host="0.0.0.0", port=8000)

    # Note: In production, you should run this behind a reverse proxy like Nginx
    # and with a proper HTTPS certificate
